@@ -179,9 +179,17 @@ class MTCNN(nn.Module):
         select_largest {bool} -- If True, if multiple faces are detected, the largest is returned.
             If False, the face with the highest detection probability is returned.
             (default: {True})
+        selection_method {string} -- Which heuristic to use for selection. Default None. If
+            specified, will override select_largest:
+                    "probability": highest probability selected
+                    "largest": largest box selected
+                    "largest_over_theshold": largest box over a certain probability selected
+                    "center_weighted_size": box size minus weighted squared offset from image center
+                (default: {None})
         keep_all {bool} -- If True, all detected faces are returned, in the order dictated by the
             select_largest parameter. If a save_path is specified, the first face is saved to that
             path and the remaining faces are saved to <save_path>1, <save_path>2 etc.
+            (default: {False})
         device {torch.device} -- The device on which to run neural net passes. Image tensors and
             models are copied to this device before running forward passes. (default: {None})
     """
@@ -189,7 +197,7 @@ class MTCNN(nn.Module):
     def __init__(
         self, image_size=160, margin=0, min_face_size=20,
         thresholds=[0.6, 0.7, 0.7], factor=0.709, post_process=True,
-        select_largest=True, keep_all=False, device=None
+        select_largest=True, selection_method=None, keep_all=False, device=None
     ):
         super().__init__()
 
@@ -201,6 +209,7 @@ class MTCNN(nn.Module):
         self.post_process = post_process
         self.select_largest = select_largest
         self.keep_all = keep_all
+        self.selection_method = selection_method
 
         self.pnet = PNet()
         self.rnet = RNet()
@@ -211,13 +220,16 @@ class MTCNN(nn.Module):
             self.device = device
             self.to(device)
 
+        if not self.selection_method:
+            self.selection_method = 'largest' if self.select_largest else 'probability'
+
     def forward(self, img, save_path=None, return_prob=False):
         """Run MTCNN face detection on a PIL image or numpy array. This method performs both
         detection and extraction of faces, returning tensors representing detected faces rather
         than the bounding boxes. To access bounding boxes, see the MTCNN.detect() method below.
         
         Arguments:
-            img {PIL.Image, np.ndarray, or list} -- A PIL image, np.ndarray, or list.
+            img {PIL.Image, np.ndarray, or list} -- A PIL image, np.ndarray, torch.Tensor, or list.
         
         Keyword Arguments:
             save_path {str} -- An optional save path for the cropped image. Note that when
@@ -243,66 +255,17 @@ class MTCNN(nn.Module):
         """
 
         # Detect faces
-        with torch.no_grad():
-            batch_boxes, batch_probs = self.detect(img)
-
-        # Determine if a batch or single image was passed
-        batch_mode = True
-        if (
-            not isinstance(img, (list, tuple)) and
-            not (isinstance(img, np.ndarray) and len(img.shape) == 4) and
-            not (isinstance(img, torch.Tensor) and len(img.shape) == 4)
-        ):
-            img = [img]
-            batch_boxes = [batch_boxes]
-            batch_probs = [batch_probs]
-            batch_mode = False
-
-        # Parse save path(s)
-        if save_path is not None:
-            if isinstance(save_path, str):
-                save_path = [save_path]
-        else:
-            save_path = [None for _ in range(len(img))]
-        
-        # Process all bounding boxes and probabilities
-        faces, probs = [], []
-        for im, box_im, prob_im, path_im in zip(img, batch_boxes, batch_probs, save_path):
-            if box_im is None:
-                faces.append(None)
-                probs.append([None] if self.keep_all else None)
-                continue
-
-            if not self.keep_all:
-                box_im = box_im[[0]]
-
-            faces_im = []
-            for i, box in enumerate(box_im):
-                face_path = path_im
-                if path_im is not None and i > 0:
-                    save_name, ext = os.path.splitext(path_im)
-                    face_path = save_name + '_' + str(i + 1) + ext
-
-                face = extract_face(im, box, self.image_size, self.margin, face_path)
-                if self.post_process:
-                    face = fixed_image_standardization(face)
-                faces_im.append(face)
-
-            if self.keep_all:
-                faces_im = torch.stack(faces_im)
-            else:
-                faces_im = faces_im[0]
-                prob_im = prob_im[0]
-            
-            faces.append(faces_im)
-            probs.append(prob_im)
-    
-        if not batch_mode:
-            faces = faces[0]
-            probs = probs[0]
+        batch_boxes, batch_probs, batch_points = self.detect(img, landmarks=True)
+        # Select faces
+        if not self.keep_all:
+            batch_boxes, batch_probs, batch_points = self.select_boxes(
+                batch_boxes, batch_probs, batch_points, img, method=self.selection_method
+            )
+        # Extract faces
+        faces = self.extract(img, batch_boxes, save_path)
 
         if return_prob:
-            return faces, probs
+            return faces, batch_probs
         else:
             return faces
 
@@ -315,7 +278,7 @@ class MTCNN(nn.Module):
         followed by the extract_face() function.
         
         Arguments:
-            img {PIL.Image, np.ndarray, or list} -- A PIL image or a list of PIL images.
+            img {PIL.Image, np.ndarray, or list} -- A PIL image, np.ndarray, torch.Tensor, or list.
 
         Keyword Arguments:
             landmarks {bool} -- Whether to return facial landmarks in addition to bounding boxes.
@@ -391,6 +354,154 @@ class MTCNN(nn.Module):
 
         return boxes, probs
 
+    def select_boxes(
+        self, all_boxes, all_probs, all_points, imgs, method='probability', threshold=0.9,
+        center_weight=2.0
+    ):
+        """Selects a single box from multiple for a given image using one of multiple heuristics.
+
+        Arguments:
+                all_boxes {np.ndarray} -- Ix0 ndarray where each element is a Nx4 ndarry of
+                    bounding boxes for N detected faces in I images (output from self.detect).
+                all_probs {np.ndarray} -- Ix0 ndarray where each element is a Nx0 ndarry of
+                    probabilities for N detected faces in I images (output from self.detect).
+                all_points {np.ndarray} -- Ix0 ndarray where each element is a Nx5x2 array of
+                    points for N detected faces. (output from self.detect).
+                imgs {PIL.Image, np.ndarray, or list} -- A PIL image, np.ndarray, torch.Tensor, or list.
+
+        Keyword Arguments:
+                method {str} -- Which heuristic to use for selection:
+                    "probability": highest probability selected
+                    "largest": largest box selected
+                    "largest_over_theshold": largest box over a certain probability selected
+                    "center_weighted_size": box size minus weighted squared offset from image center
+                    (default: {'probability'})
+                threshold {float} -- theshold for "largest_over_threshold" method. (default: {0.9})
+                center_weight {float} -- weight for squared offset in center weighted size method.
+                    (default: {2.0})
+
+        Returns:
+                tuple(numpy.ndarray, numpy.ndarray, numpy.ndarray) -- nx4 ndarray of bounding boxes
+                    for n images. Ix0 array of probabilities for each box, array of landmark points.
+        """
+
+        #copying batch detection from extract, but would be easier to ensure detect creates consistent output.
+        batch_mode = True
+        if (
+                not isinstance(imgs, (list, tuple)) and
+                not (isinstance(imgs, np.ndarray) and len(imgs.shape) == 4) and
+                not (isinstance(imgs, torch.Tensor) and len(imgs.shape) == 4)
+        ):
+            imgs = [imgs]
+            all_boxes = [all_boxes]
+            all_probs = [all_probs]
+            all_points = [all_points]
+            batch_mode = False
+
+        selected_boxes, selected_probs, selected_points = [], [], []
+        for boxes, points, probs, img in zip(all_boxes, all_points, all_probs, imgs):
+
+            boxes = np.array(boxes)
+            probs = np.array(probs)
+            points = np.array(points)
+
+            if len(boxes) == 0:
+                selected_boxes.append(None)
+                selected_probs.append([None])
+                selected_points.append(None)
+                continue
+            elif method == 'largest':
+                box_order = np.argsort((boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1]))[::-1]
+            elif method == 'probability':
+                box_order = np.argsort(probs)[::-1]
+            elif method == 'center_weighted_size':
+                box_sizes = (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])
+                img_center = (img.width / 2, img.height/2)
+                box_centers = np.array(list(zip((boxes[:, 0] + boxes[:, 2]) / 2, (boxes[:, 1] + boxes[:, 3]) / 2)))
+                offsets = box_centers - img_center
+                offset_dist_squared = np.sum(np.power(offsets, 2.0), 1)
+                box_order = np.argsort(box_sizes - offset_dist_squared * center_weight)[::-1]
+            elif method == 'largest_over_threshold':
+                box_mask = probs > threshold
+                boxes = boxes[box_mask]
+                box_order = np.argsort((boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1]))[::-1]
+                if sum(box_mask) == 0:
+                    selected_boxes.append(None)
+                    selected_probs.append([None])
+                    selected_points.append(None)
+                    continue
+
+            box = boxes[box_order][[0]]
+            prob = probs[box_order][[0]]
+            point = points[box_order][[0]]
+            selected_boxes.append(box)
+            selected_probs.append(prob)
+            selected_points.append(point)
+
+        if batch_mode:
+            selected_boxes = np.array(selected_boxes)
+            selected_probs = np.array(selected_probs)
+            selected_points = np.array(selected_points)
+        else:
+            selected_boxes = selected_boxes[0]
+            selected_probs = selected_probs[0][0]
+            selected_points = selected_points[0]
+
+        return selected_boxes, selected_probs, selected_points
+
+    def extract(self, img, batch_boxes, save_path):
+        # Determine if a batch or single image was passed
+        batch_mode = True
+        if (
+                not isinstance(img, (list, tuple)) and
+                not (isinstance(img, np.ndarray) and len(img.shape) == 4) and
+                not (isinstance(img, torch.Tensor) and len(img.shape) == 4)
+        ):
+            img = [img]
+            batch_boxes = [batch_boxes]
+            batch_mode = False
+
+        # Parse save path(s)
+        if save_path is not None:
+            if isinstance(save_path, str):
+                save_path = [save_path]
+        else:
+            save_path = [None for _ in range(len(img))]
+
+        # Process all bounding boxes
+        faces = []
+        for im, box_im, path_im in zip(img, batch_boxes, save_path):
+            if box_im is None:
+                faces.append(None)
+                continue
+
+            if not self.keep_all:
+                box_im = box_im[[0]]
+
+            faces_im = []
+            for i, box in enumerate(box_im):
+                face_path = path_im
+                if path_im is not None and i > 0:
+                    save_name, ext = os.path.splitext(path_im)
+                    face_path = save_name + '_' + str(i + 1) + ext
+
+                face = extract_face(im, box, self.image_size, self.margin, face_path)
+                if self.post_process:
+                    face = fixed_image_standardization(face)
+                faces_im.append(face)
+
+            if self.keep_all:
+                faces_im = torch.stack(faces_im)
+            else:
+                faces_im = faces_im[0]
+
+            faces.append(faces_im)
+
+        if not batch_mode:
+            faces = faces[0]
+
+        return faces
+
 
 def fixed_image_standardization(image_tensor):
     processed_tensor = (image_tensor - 127.5) / 128.0
@@ -403,3 +514,4 @@ def prewhiten(x):
     std_adj = std.clamp(min=1.0/(float(x.numel())**0.5))
     y = (x - mean) / std_adj
     return y
+
